[description]

The virt-fw-dump utility prints the structure and content of edk2
firmware volumes to stdout.

[examples]

virt-fw-dump --input /usr/share/edk2/ovmf/OVMF_CODE.fd

[author]

Gerd Hoffmann <kraxel@redhat.com>

